# -*- coding: utf-8 -*-
"""
Created on Thu Feb 23 16:23:53 2023

@author: qq102
"""
import pandas as pd
import numpy as np
import re
from functools import reduce
import seaborn as sns
import matplotlib.pyplot as plt
import shap

from packaging import version
import warnings

try:
    import sklearn
    sklearn_exists = True
except ImportError:
    sklearn_exists = False
    
if sklearn_exists:
    if version.parse(sklearn.__version__) < version.parse('1.0.2'):
        warnings.warn("Your sklearn version is older than 1.0.2  +\
                      and may be not operation correct.")
    from sklearn.preprocessing import MinMaxScaler
else:
    warnings.warn("Sklearn not found. MinMaxScaler will not be available.") 

# from sklearn.preprocessing import MinMaxScaler




##库核心框架
class SHAP_importance_sum:
    def __init__(self):
        self.shapley_data_all=None
    
    @staticmethod
    def str_contain(str_input,input_names):
        pattern=re.compile(str_input)
        str_outcom=[]
        for i in input_names:
            if re.search(pattern,i):
                str_outcom.append(i)
        return(str_outcom)
    
    @staticmethod
    def feature_shap_statistic(SHAP_data_all,depleted_ID_len=0):
        SHAP_data_all_feature_shap=list(map(lambda x:SHAP_data_all[x].iloc[:,:len(SHAP_data_all[x].columns)-depleted_ID_len],SHAP_data_all.keys()))
        SHAP_data_all_feature_shap=dict(zip(list(SHAP_data_all.keys()),SHAP_data_all_feature_shap))
        for i,value in enumerate(SHAP_data_all_feature_shap):
            feature_inner_name=SHAP_data_all_feature_shap[value].columns
            SHAP_data_all_feature_shap[value]=list(map(lambda x:pd.DataFrame(SHAP_data_all_feature_shap[value][x]),SHAP_data_all_feature_shap[value].columns))
            SHAP_data_all_feature_shap[value]=dict(zip(list(feature_inner_name),SHAP_data_all_feature_shap[value]))
            for i_1,value_1 in enumerate(SHAP_data_all_feature_shap[value]):
                SHAP_data_all_feature_shap[value][value_1].columns=[value]
        SHAP_data_all_feature_shap_outcome=dict()
        
        for i in list(feature_inner_name):
            SHAP_data_all_feature_shap_outcome[i]=list(map(lambda x: SHAP_data_all_feature_shap[x][i],SHAP_data_all_feature_shap.keys()))   
              
            SHAP_data_all_feature_shap_outcome[i]=reduce(lambda x,y: pd.concat([x,y],axis=1),SHAP_data_all_feature_shap_outcome[i])       
            
        return(SHAP_data_all_feature_shap_outcome)
    
    def IG_importance_sum_claulate_process(self,depleted_ID_len=0,file_name='human_'):
        scale=MinMaxScaler()
        self.type_data_shap_sum_all=dict()
        self.type_data_shap_sum_all_outcome=dict()
        self.shap_feature_importance=dict()
        self.shap_feature_importance_T=None
        for value in self.shapley_data_all.keys():
            ##shap求和，并且归一化
            
            self.type_data_shap_sum_all[value]=np.array(self.shapley_data_all[value].iloc[:,:len(self.shapley_data_all[value].columns)-depleted_ID_len]).sum(axis=1)
            self.type_data_shap_sum_all[value]=pd.DataFrame({'sum_shap':self.type_data_shap_sum_all[value].reshape(-1,),
                                                        'scale':scale.fit_transform(self.type_data_shap_sum_all[value].reshape(-1,1)).reshape(-1,)},
                                                        index=self.shapley_data_all[value].index)
            self.type_data_shap_sum_all_outcome[value]=pd.merge(self.shapley_data_all[value],self.type_data_shap_sum_all[value],how="inner",left_index=True,right_index=True)
            
            
            ##特征的shap绝对值相加作为特征的重要性
            self.shap_feature_importance[value]=np.abs(self.shapley_data_all[value].iloc[:,:len(self.shapley_data_all[value].columns)-depleted_ID_len])
            self.shap_feature_importance[value]=self.shap_feature_importance[value].mean(axis=0)
            self.shap_feature_importance[value]=pd.DataFrame(self.shap_feature_importance[value],columns=[value])

        self.shap_feature_importance=reduce(lambda x,y: pd.merge(x,y,how="inner",left_index=True,right_index=True),self.shap_feature_importance.values())

        self.shap_feature_importance.columns=[re.compile('('+file_name+')|(_shap_value)').sub('',name) for name in self.shap_feature_importance.columns]
        ##shap_feature_importance转置
        self.shap_feature_importance_T=self.shap_feature_importance.T
        return{'type_data_shap_sum_all_outcome':self.type_data_shap_sum_all_outcome,
                'shap_feature_importance':self.shap_feature_importance,
                'shap_feature_importance_T':self.shap_feature_importance_T}
    
    
    def SHAP_importance_sum_claulate_process(self, base_probablity_dict, depleted_ID_len=0, file_name='human_'):
        scale = MinMaxScaler()
        self.type_data_shap_sum_all = dict()
        self.type_data_shap_sum_all_outcome = dict()
        self.shap_feature_importance = dict()
        self.shap_feature_importance_T = None
        for value in self.shapley_data_all.keys():
            # shap求和，并且归一化
            sum_shap = np.array(self.shapley_data_all[value].iloc[:, :len(self.shapley_data_all[value].columns)-depleted_ID_len]).sum(axis=1)

            # 获取对应的基准概率
            base_value = base_probablity_dict[value]

            # 创建 sum_shap 的副本，用于计算 probablity
            probablity = sum_shap.copy()

            # 将基准概率加到 probablity 值上
            probablity += base_value

            self.type_data_shap_sum_all[value] = pd.DataFrame({'sum_shap': sum_shap.reshape(-1,),
                                                               'scale': scale.fit_transform(sum_shap.reshape(-1,1)).reshape(-1,),
                                                               'probablity': probablity},  # 新增probablity列
                                                              index=self.shapley_data_all[value].index)
            self.type_data_shap_sum_all_outcome[value] = pd.merge(self.shapley_data_all[value], self.type_data_shap_sum_all[value], how="inner", left_index=True, right_index=True)

            # 特征的shap绝对值相加作为特征的重要性
            self.shap_feature_importance[value] = np.abs(self.shapley_data_all[value].iloc[:, :len(self.shapley_data_all[value].columns)-depleted_ID_len])
            self.shap_feature_importance[value] = self.shap_feature_importance[value].mean(axis=0)
            self.shap_feature_importance[value] = pd.DataFrame(self.shap_feature_importance[value], columns=[value])

        self.shap_feature_importance = reduce(lambda x, y: pd.merge(x, y, how="inner", left_index=True, right_index=True), self.shap_feature_importance.values())

        self.shap_feature_importance.columns = [re.compile('('+file_name+')|(_shap_value)').sub('', name) for name in self.shap_feature_importance.columns]
        # shap_feature_importance转置
        self.shap_feature_importance_T = self.shap_feature_importance.T
        return {'type_data_shap_sum_all_outcome': self.type_data_shap_sum_all_outcome,
                'shap_feature_importance': self.shap_feature_importance,
                'shap_feature_importance_T': self.shap_feature_importance_T}
    
    def calculate_feature_shap_bins(self, keys, feature, bins_num=10, q_num=10):
        result = pd.DataFrame()
        for key in keys:
            data = self.shapley_data_all[key]
            feature_data = data[feature]
            # 先按分位数分箱
            quantiles = pd.qcut(feature_data, q=q_num, duplicates='drop')
            # 记录全局箱子编号
            global_bins = 0
            bins_list = []
            # 在每个分位数区间内按等距分箱
            for _, group in feature_data.groupby(quantiles):
                # 由于group的index是feature_data的index，所以可以通过这个index从feature_data中取到对应的值
                group_values = feature_data[group.index]
                bins = pd.cut(group_values, bins=bins_num, include_lowest=True)
                # 获取箱子编号，并加上全局箱子编号
                bins_codes = bins.cat.codes + global_bins
                bins_list.append(bins_codes)
                # 更新全局箱子编号
                global_bins = bins_codes.max() + 1
            # 拼接所有的箱子编号
            bins_codes = pd.concat(bins_list)
            temp_df = pd.DataFrame({
                f'shap_values_{key}': feature_data,
                f'bins_{key}': bins_codes
            })
            if result.empty:
                result = temp_df
            else:
                result = result.join(temp_df, how='outer')
        return result
    


    

class FeaturePlotSource:
    def __init__(self, X_data, shapley_data_all):
        self.X_data = X_data
        self.shapley_data_all = shapley_data_all
    
    def plot(self):
        pass


class FeaturePlot(FeaturePlotSource):

    class ShapImportancePlot:
        def __init__(self, func):
            self.func = func
            self.self_obj = None
        
        def __get__(self, instance, owner):
            self.self_obj = instance
            return self
        
        def __call__(self, *args, **kwargs):
            self.plot(*args, **kwargs)
        
        def plot(self, file_name, save_path, plot_size, *args, **kwargs):
            shap_values_plot = [value for value in self.self_obj.shapley_data_all.values()]
            shap_values_plot = [value.to_numpy() if not isinstance(value, np.ndarray) 
                                else value for value in shap_values_plot]
            shap_values_class = [name for name in self.self_obj.shapley_data_all.keys()]
            
            plot_shap_importance_outcom = plt.figure(figsize=plot_size, dpi=1000)
            plt.title(file_name + '_shap_importance')
            shap.summary_plot(shap_values_plot, self.self_obj.X_data, plot_type="bar", class_names=shap_values_class, 
                              show=False, plot_size=plot_size, *args, **kwargs)
            plot_shap_importance_outcom.savefig(save_path + file_name + '_shap_importance.png', dpi=1000, bbox_inches="tight")
            plot_shap_importance_outcom.savefig(save_path + file_name + '_shap_importance.pdf', dpi=1000, bbox_inches="tight")
            plt.close()

    @ShapImportancePlot
    def shap_importance_plot(self, file_name, save_path, plot_size, *args, **kwargs):
        pass


    class FeatureSummaryPlot:
        def __init__(self, func):
            self.func = func
            self.self_obj = None
        
        def __get__(self, instance, owner):
            self.self_obj = instance
            return self
        def __call__(self, *args, **kwargs):
            self.plot(*args, **kwargs)
        def shap_summary_plot(self, shapley_data_all, X_input, class_names, *args, **kwargs):
            plot_shap_summary_outcom = plt.figure(figsize=self.plot_size, dpi=1000)
            plt.title(self.file_name + "_" + class_names + '_shap_summary')
            shap.summary_plot(shapley_data_all, X_input, plot_type="dot", show=False, plot_size=self.plot_size, *args, **kwargs)
            plot_shap_summary_outcom.savefig(self.save_path + class_names + self.file_name + "_shap_summary.png", dpi=1000, bbox_inches="tight")
            plot_shap_summary_outcom.savefig(self.save_path + class_names + self.file_name + "_shap_summary.pdf", dpi=1000, bbox_inches="tight")
            plt.close()

        def plot(self, file_name, save_path,plot_size,*args, **kwargs):
            self.file_name = file_name
            self.save_path = save_path
            self.plot_size = plot_size
            shap_values_plot = [value for value in self.self_obj.shapley_data_all.values()]
            shap_values_plot = [value.to_numpy() if not isinstance(value, np.ndarray) 
                                else value for value in shap_values_plot]
            shap_values_class = [name for name in self.self_obj.shapley_data_all.keys()]
            for i, value in enumerate(shap_values_class):
                self.shap_summary_plot(shapley_data_all=shap_values_plot[i], X_input=self.self_obj.X_data,
                                        class_names=value, *args, **kwargs)
    
    @FeatureSummaryPlot
    def Shapley_summary_plot(self, file_name, save_path, plot_size, *args, **kwargs):
        pass
    
    class FeatureHistPlot:
        def __init__(self, func):
            self.func = func
            self.self_obj = None
        
        def __get__(self, instance, owner):
            self.self_obj = instance
            return self
        def __call__(self, *args, **kwargs):
            self.plot(*args, **kwargs)
        def plot(self,shap_indicate_feature,save_path, file_name, png_plot=True,
                 jointplot_kwargs=None, marg_x_kwargs=None, marg_y_kwargs=None):
            self.shap_indicate_feature = shap_indicate_feature
            self.save_path = save_path
            self.file_name = file_name
            self.png_plot = png_plot
            
            if jointplot_kwargs is None:
                jointplot_kwargs = {}
            if marg_x_kwargs is None:
                marg_x_kwargs = {}
            if marg_y_kwargs is None:
                marg_y_kwargs = {}
                
            for value in list(self.self_obj.shapley_data_all.keys()):
                for i_name in self.shap_indicate_feature: 
                    df = pd.DataFrame({'Feature_value':self.self_obj.X_data[i_name],
                                       'shapley_value':self.self_obj.shapley_data_all[value][i_name]})
                    ax = sns.jointplot(x=df['Feature_value'], y=df['shapley_value'],
                                       kind='hist', color='r',**jointplot_kwargs)
                    
                    ax.set_axis_labels(xlabel=f"Feature value of {i_name}",
                                       ylabel='shapley_value')
                    # # 设置直方图bin的数量
                    # ax.ax_marg_x.hist(df['Feature_value'], **marg_x_kwargs)
                    # ax.ax_marg_y.hist(df['shapley_value'], **marg_y_kwargs,orientation='horizontal')
                    
                    # 使用 Matplotlib 绘制坐标轴上的直方图
                    fig = plt.gcf()
                    ax1 = fig.add_axes(ax.ax_marg_x.get_position())
                    ax1.hist(df['Feature_value'], **marg_x_kwargs)
                    ax1.set_xticks([])
                    ax2 = fig.add_axes(ax.ax_marg_y.get_position())
                    ax2.hist(df['shapley_value'], orientation='horizontal',**marg_y_kwargs)
                    ax2.set_yticks([])
                    fig.suptitle(f"{value} {self.file_name}",y=1.05)
                    ax.savefig(self.save_path + 
                               f"{self.file_name}_{value}_{i_name}_histplot.pdf",
                               dpi=1000, bbox_inches="tight")
                    if self.png_plot:
                        ax.savefig(self.save_path + 
                                   f"{self.file_name}_{value}_{i_name}_histplot.png",
                                   dpi=1000, bbox_inches="tight")
                    plt.close(ax.figure)
    @FeatureHistPlot
    def feature_hist_plot(self,shap_indicate_feature,save_path,
                                file_name, png_plot,*args, **kwargs):
        pass
    
    
    
    
    class FeatureScatterplot(FeaturePlotSource):
        def __init__(self, func):
            self.func = func
            self.self_obj = None
        
        def __get__(self, instance, owner):
            self.self_obj = instance
            return self
        def __call__(self, *args, **kwargs):
            self.plot(*args, **kwargs)
            
        def plot(self,shap_indicate_feature,save_path,file_name='human_train',png_plot=True,dependence=True,*args, **kwargs):
            self.shap_indicate_feature = shap_indicate_feature
            self.save_path = save_path
            self.file_name = file_name
            self.png_plot = png_plot
            for value in list(self.self_obj.shapley_data_all.keys()):
                for i_name in self.shap_indicate_feature:
                    plt.figure()
                    if dependence:
                        shap.dependence_plot(i_name.split('*_')[0], np.array(self.self_obj.shapley_data_all[value]), 
                                             self.self_obj.X_data, title=self.file_name+"_"+value+"_"+'shap_dependence', 
                                             show=False, interaction_index=i_name.split('*_')[1],*args, **kwargs)
                        plt.savefig(self.save_path+self.file_name+"_"+value+"_"+i_name.split('*_')[0]+'vs_'+i_name.split('*_')[1]+"_shap_dependence.pdf",dpi=1000,bbox_inches="tight")
                        if self.png_plot:
                            plt.savefig(self.save_path+self.file_name+"_"+value+"_"+i_name.split('*_')[0]+'vs_'+i_name.split('*_')[1]+"_shap_dependence.png",dpi=1000,bbox_inches="tight")
                    else:
                        shap.dependence_plot(i_name, np.array(self.self_obj.shapley_data_all[value]), self.self_obj.X_data,
                                             title=self.file_name+"_"+value+"_"+'shap_independence', show=False,
                                             interaction_index=None,*args, **kwargs)
                        plt.savefig(self.save_path+self.file_name+"_"+value+"_"+i_name+"_shap_nodependence.pdf",dpi=1000,bbox_inches="tight")
                        if self.png_plot:
                            plt.savefig(self.save_path+self.file_name+"_"+value+"_"+i_name+"_shap_nodependence.png",dpi=1000,bbox_inches="tight")
                    plt.close()

    @FeatureScatterplot
    def feature_scatter_plot(self,shap_indicate_feature,save_path,
                                file_name, png_plot,dependence,*args, **kwargs):
        pass        
        

   

class LocalAnalysisPlot(FeaturePlotSource):
    def __init__(self, X_data, shapley_data_all,probability_data):
        super().__init__(X_data, shapley_data_all)
        self.probability_data = probability_data

     
    def plot(self,protein_indicate,save_path, file_name='huamn', plot_size=(10,10),png_plot=True,
             feature_remain=5,positive_color='#FFB6C1',negative_color='#ADD8E6'):
        # protein_indicate= protein_indicate
        # save_path = save_path
        # file_name = file_name
        # png_plot = png_plot
        
            
        for value in list(self.shapley_data_all.keys()):
            for i_name in protein_indicate:
                ##创建绘图概率
                probability_predict=self.probability_data[value].loc[i_name].to_numpy()[0]
                
                ##创建绘图列表
                df = pd.DataFrame({'Feature_value':self.X_data.loc[i_name],
                                   'shapley_value':self.shapley_data_all[value].loc[i_name]})
                df['shapley_value_abs'] = df['shapley_value'].abs()
                df= df.sort_values(by='shapley_value_abs', ascending=False).iloc[:feature_remain]
                df['abs_rank'] = df['shapley_value_abs'].rank( ascending=True)
                feature_name = df.index.tolist()
                feature_value=df['Feature_value'].tolist()
                yticklabels = [f"{feature_name[i]} ({feature_value[i]:.4f})" for i in range(len(feature_name))]
                # 将正数和负数数据分别存储到两个不同的数组中
                # 使用条件语句将正数和负数数据分别存储到两个不同的DataFrame中
                positive_data = df[df['shapley_value'] > 0]
                negative_data = df[df['shapley_value'] <= 0]
                
                # 创建一个包含两个水平直方图的子图
                fig, ax = plt.subplots()

                # 绘制正数数据的水平直方图
                ax.barh(positive_data['abs_rank'].to_numpy(), 
                        positive_data['shapley_value'], align='center', color=positive_color)

                # 绘制负数数据的水平直方图
                ax.barh(negative_data['abs_rank'].to_numpy(), 
                        negative_data['shapley_value'], align='center', color=negative_color)
                ax.set_yticks(df['abs_rank'].to_numpy())
                ax.set_yticklabels(yticklabels)
                
                # 添加概率预测值的注释
                ax.text(1, 1, f"{i_name} for {value} Probability Prediction: {probability_predict:.2f}", 
                transform=ax.transAxes,
                verticalalignment='bottom', horizontalalignment='right')
                ax.set_xlabel('shapley_value')
                ax.set_ylabel(f"Feature value of {i_name}")

                fig.set_size_inches(plot_size)
                fig.savefig(save_path + 
                           f"{file_name}_{value}_{i_name}_localplot.pdf",
                           dpi=1000, bbox_inches="tight")
                if png_plot:
                    fig.savefig(save_path + 
                               f"{file_name}_{value}_{i_name}_localplot.png",
                               dpi=1000, bbox_inches="tight")
                plt.close(ax.figure)



   
class Featureinteractionplot(FeaturePlotSource):
    def plot(self, interaction=True):
        for value in list(self.shapley_data_all.keys()):
            for i_name in self.shap_indicate_feature:
                if interaction:
                    shap.dependence_plot((i_name.split('*_')[0],i_name.split('*_')[1]), np.array(self.shapley_data_all[value]), 
                                         self.X_data , 
                                         show=False, interaction_index=None)
                    plt.title(self.file_name+"_"+value+"_"+'interaction_deprincipal')
                    plt.savefig(self.save_path+self.file_name+"_"+value+"_"+i_name.split('*_')[0]+'vs_'+i_name.split('*_')[1]+"_shap_interaction_depc.pdf",dpi=1000,bbox_inches="tight")
                    if self.png_plot:
                        plt.savefig(self.save_path+self.file_name+"_"+value+"_"+i_name.split('*_')[0]+'vs_'+i_name.split('*_')[1]+"_shap_interaction_depc.png",dpi=1000,bbox_inches="tight")
                else:
                    shap.dependence_plot((i_name,i_name), np.array(self.shapley_data_all[value]), self.X_data,
                                          show=False,
                                         interaction_index=None)
                    plt.title(self.file_name+"_"+value+"_"+'pf_residuals')
                    plt.savefig(self.save_path+self.file_name+"_"+value+"_"+i_name+"_shap_residuals.pdf",dpi=1000,bbox_inches="tight")
                    if self.png_plot:
                        plt.savefig(self.save_path+self.file_name+"_"+value+"_"+i_name+"_shap_residuals.png",dpi=1000,bbox_inches="tight")
                plt.close()

    def interaction_deprincipal_plot(self,shap_indicate_feature,save_path, file_name='human_train', png_plot=True):
        self.shap_indicate_feature = shap_indicate_feature
        self.save_path = save_path
        self.file_name = file_name
        self.png_plot = png_plot
        self.plot(interaction=True)
        
    def principal_feature_residuals_plot(self,shap_indicate_feature,save_path, file_name='human_train', png_plot=True):
        self.shap_indicate_feature = shap_indicate_feature
        self.save_path = save_path
        self.file_name = file_name
        self.png_plot = png_plot
        self.plot(interaction=False)

    def interaction_all_plot(self, file_name, save_path,max_num_display=25):
        for name in self.shapley_data_all.keys():
            tmp = np.abs(self.shapley_data_all[name]).sum(0)/len(self.X_data.index)
            for i in range(tmp.shape[0]):
                tmp[i,i] = 0
            inds = np.argsort(-tmp.sum(0))[:max_num_display]##最多只展示25个特征
            tmp2 = tmp[inds,:][:,inds]
            plt.figure(figsize=(12,12))
            plt.title(name+file_name+'_interaction')
            plt.imshow(tmp2)
            plt.yticks(range(tmp2.shape[0]), self.X_data.columns[inds], rotation=50.4, horizontalalignment="right")
            plt.xticks(range(tmp2.shape[0]), self.X_data.columns[inds], rotation=50.4, horizontalalignment="left")
            plt.gca().xaxis.tick_top()
            plt.colorbar(plt.imshow(tmp2))
            plt.savefig(save_path+file_name+"_"+name+"_"+"shap_interaction_all.pdf",dpi=1000,bbox_inches="tight")
            plt.close()


class FeatureinteractionIndexsave(FeaturePlotSource):
    
    def calculate_interaction_feature_index(self):
        self.shap_interaction_feature_index = {}
        for i in self.shapley_data_all.keys():
            shap_train_interaction_feature_index = list(map(lambda x: pd.DataFrame(2*self.shapley_data_all[i][:,x,:],
                                                                           index=self.X_data.index,columns=self.X_data.columns),
                                                  range(len(self.X_data.columns))))
            shap_train_interaction_feature_index = dict(zip(list(self.X_data.columns), shap_train_interaction_feature_index)) 
            self.shap_interaction_feature_index[i] = shap_train_interaction_feature_index
    
    def save_interaction_feature_index(self,shap_indicate_feature,save_path,file_name):
        self.shap_indicate_feature=shap_indicate_feature
        for i in self.shap_interaction_feature_index.keys():
            for value in self.shap_indicate_feature:
                self.shap_interaction_feature_index[i][value].to_csv(save_path+file_name+'_'+i+'_'+value+'_shap_interaction_feature_index.csv')


def predict_probability_Fcombine(cluster_data_all, type_data_all, cluster_sum_shap, gene_ID, type_probability):
    # Initialize dictionaries to store results
    cluster_sum_shap_beyond_zero_outcome = {}
    cluster_sum_shap_less_zero_outcome = {}
    
    # Calculate cluster's shap value for proteins with values greater than zero
    for key, value in cluster_data_all.items():
        cluster_sum_shap_beyond_zero = value[value[cluster_sum_shap] > 0]
        new_df_merge = None
        for key2, value2 in type_data_all.items():
            # Merge the two data frames on the 'gene_ID' column using an inner join
            merged = pd.merge(cluster_sum_shap_beyond_zero, value2, on=gene_ID, how='inner')
            # Create a new data frame with the mean probability value, indexed by 'key2'
            new_df = pd.DataFrame({key:[merged[type_probability].mean()]},index=[key2])
            new_df_merge = pd.concat([new_df_merge, new_df], axis=0)
        # Append the new data frame to the dictionary
        cluster_sum_shap_beyond_zero_outcome[key] = new_df_merge
    # Concatenate all dataframes in the dictionary
    cluster_sum_shap_beyond_zero_outcome = reduce(lambda x,y: pd.concat([x, y], axis=1),cluster_sum_shap_beyond_zero_outcome.values())
    
    # Calculate cluster's shap value for proteins with values less than zero
    for key, value in cluster_data_all.items():
        cluster_sum_shap_less_zero = value[value[cluster_sum_shap] <= 0]
        new_df_merge = None
        for key2, value2 in type_data_all.items():
            # Merge the two data frames on the 'gene_ID' column using an inner join
            merged = pd.merge(cluster_sum_shap_less_zero, value2, on=gene_ID, how='inner')
            # Create a new data frame with the mean probability value, indexed by 'key2'
            new_df = pd.DataFrame({key:[merged[type_probability].mean()]},index=[key2])
            new_df_merge = pd.concat([new_df_merge, new_df], axis=0)
        # Append the new data frame to the dictionary
        cluster_sum_shap_less_zero_outcome[key] = new_df_merge
    # Concatenate all dataframes in the dictionary
    cluster_sum_shap_less_zero_outcome = reduce(lambda x,y: pd.concat([x, y], axis=1),cluster_sum_shap_less_zero_outcome.values())
    
    return cluster_sum_shap_beyond_zero_outcome, cluster_sum_shap_less_zero_outcome



def cluster_data_divided_feature(cluster_data_all,  cluster_indicate, cluster_sum_shap, gene_ID):

    # create a list of dataframes and reassign to cluster_data_all
    cluster_data_all_list = [cluster_data_all[x].loc[:, [gene_ID]+cluster_indicate] for x in cluster_data_all.keys()]
    cluster_data_all = dict(zip(cluster_data_all.keys(), cluster_data_all_list))
    
    # modify column names of dataframes
    for i_cluster in cluster_data_all.keys():
        for x in cluster_indicate:
            cluster_data_all[i_cluster].rename(columns={x: f"{x}_{i_cluster}"}, inplace=True)
        
        # group dataframes by columns
        grouped = cluster_data_all[i_cluster].groupby(cluster_data_all[i_cluster].columns, axis=1)

        new_dict = {}
        for col in grouped.groups:
            sub_df = grouped.get_group(col)
            new_dict[col] = sub_df
            new_dict[col]
            new_dict[col].insert(0, 'new_'+gene_ID, cluster_data_all[i_cluster][gene_ID])
            new_dict.pop(gene_ID, None)
            
            for key in new_dict.keys():
                # change the column name from 'gene_ID' to 'ID'
                new_dict[key] = new_dict[key].rename(columns={'new_'+gene_ID: gene_ID, new_dict[key].columns[-1]: cluster_sum_shap})
                
        cluster_data_all[i_cluster] = new_dict
        
    return cluster_data_all



def predict_probability_Fsingle(cluster_data_all, type_data_all, type_probability, cluster_sum_shap,feature_indicate,gene_ID):
    # Join words in a pattern
    pattern = '|'.join(feature_indicate)
    # Process data and create a new dataframe for each cluster_sum_shap_data
    def process_data(x1,key2,i_type):
        new_df = pd.DataFrame({key2:[x1.merge(type_data_all[i_type], on=gene_ID,how='inner' )[type_probability].mean()]},index=[i_type])
        return new_df
    
    
    # Filter cluster data where cluster_sum_shap is greater than or equal to zero
    cluster_sum_shap_beyond_zero = {k1: {k2: v2[v2[cluster_sum_shap] > 0] for k2, v2 in v1.items()} 
                                  for k1, v1 in cluster_data_all.items()}
    
    # Concatenate dataframes for each type and cluster_sum_shap_beyond_zero
    cluster_sum_shap_beyond_zero = {k1: {k2: pd.concat([process_data(v2,k2,i_type) for i_type in type_data_all.keys()],axis=0)  
                                      for k2, v2 in v1.items()} 
                                  for k1, v1 in cluster_sum_shap_beyond_zero.items()}
    
    # Rename keys using regex pattern
    cluster_sum_shap_beyond_zero_outcome = {k1:{re.findall(pattern,k2)[0]: v2 for k2 , v2 in v1.items()} 
                                          for k1, v1 in cluster_sum_shap_beyond_zero.items()}
    
    # Concatenate dataframes for each cluster_sum_shap
    cluster_sum_shap_beyond_zero_outcome = {k1: pd.concat([v2[k1] for k2 , v2 in cluster_sum_shap_beyond_zero_outcome.items() ],axis=1) 
                                          for i1, k1 in enumerate(feature_indicate)}

    # Filter cluster data where cluster_sum_shap is less than or equal to zero
    cluster_sum_shap_less_zero = {k1: {k2: v2[v2[cluster_sum_shap] <= 0] for k2, v2 in v1.items()} 
                                  for k1, v1 in cluster_data_all.items()}
    
    # Concatenate dataframes for each type and cluster_sum_shap_less_zero
    cluster_sum_shap_less_zero = {k1: {k2: pd.concat([process_data(v2,k2,i_type) for i_type in type_data_all.keys()],axis=0)  
                                      for k2, v2 in v1.items()} 
                                  for k1, v1 in cluster_sum_shap_less_zero.items()}
    
    # Rename keys using regex pattern
    cluster_sum_shap_less_zero_outcome = {k1:{re.findall(pattern,k2)[0]: v2 for k2 , v2 in v1.items()} 
                                          for k1, v1 in cluster_sum_shap_less_zero.items()}
    
    # Concatenate dataframes for each cluster_sum_shap
    cluster_sum_shap_less_zero_outcome = {k1: pd.concat([v2[k1] for k2 , v2 in cluster_sum_shap_less_zero_outcome.items() ],axis=1) 
                                          for i1, k1 in enumerate(feature_indicate)}
    return {'cluster_sum_shap_beyond_zero_outcome':cluster_sum_shap_beyond_zero_outcome, 'cluster_sum_shap_less_zero_outcome':cluster_sum_shap_less_zero_outcome}



    
 






import numpy as np
import pandas as pd

try:
    import torch
    from captum.attr import IntegratedGradients
    from captum.attr import DeepLift
except ImportError:
    raise ImportError("Missing dependencies: Please install 'torch' and 'captum' manually.")




class IntegratedGradientsCalculator:

    def __init__(self, model, X_input, X_input_tensor,y_input,batch_size=100, n_steps=50):
        self.model = model
        self.X_input = X_input 
        self.X_input_tensor=X_input_tensor
        self.y_input = y_input
        self.batch_size=batch_size
        self.n_steps=n_steps
    def _compute_integrated_gradients_define(self,target_type):
        
        device = self.X_input_tensor.device  
        baseline = torch.zeros(self.X_input_tensor.shape).to(device)
        
        ig = IntegratedGradients(self.model)

        all_attributions = []
        all_deltas = []

        num_batches = (self.X_input_tensor.shape[0] + self.batch_size - 1) // self.batch_size

        for batch_idx in range(num_batches):
            start_idx = batch_idx * self.batch_size
            end_idx = min((batch_idx + 1) * self.batch_size, self.X_input_tensor.shape[0])
            X_batch = self.X_input_tensor[start_idx:end_idx]

            attributions, delta = ig.attribute(X_batch, baseline[start_idx:end_idx], 
                            target=target_type, return_convergence_delta=True,
                            n_steps=self.n_steps)

            attributions_np = attributions.cpu().detach().numpy()
            delta_np = delta.cpu().detach().numpy()

            all_attributions.append(attributions_np)
            all_deltas.append(delta_np)

            torch.cuda.empty_cache()
            torch.cuda.synchronize()

        all_attributions = np.concatenate(all_attributions, axis=0)
        all_deltas = np.concatenate(all_deltas, axis=0)
        
        return all_attributions, all_deltas
    
    def compute_integrated_gradients(self,type_class):
        
        self.all_attributions_values=[]
        self.all_deltas=[]
        for i in range(len(type_class)):
            all_attributions_values_inn, all_deltas_inn=self._compute_integrated_gradients_define(i)
            self.all_attributions_values.append(all_attributions_values_inn)
            self.all_deltas.append(all_deltas_inn)
        # self.all_attributions_values, self.all_deltas=list(map
        #                     (lambda x: self._compute_integrated_gradients_define(x),range(len(type_class))))
        
        self.all_attributions_values=list(map(lambda x: pd.DataFrame(x,index=self.X_input.index,columns=self.X_input.columns),self.all_attributions_values))
        self.all_deltas=list(map(lambda x: pd.DataFrame(x,index=self.X_input.index,columns=['approximation error']),self.all_deltas))
       
        self.all_attributions_values=dict(zip(type_class,self.all_attributions_values))
        self.all_deltas=dict(zip(type_class,self.all_deltas))
        
        return self.all_attributions_values, self.all_deltas
    def integrated_gradients_save(self,X_predict_input,save_path):
        all_attributions_value_save = {key: value.join(self.y_input, how="inner") for key,value in self.all_attributions_values.items()}
        all_attributions_value_save = {key: value.join(X_predict_input, how="inner") for key, value in all_attributions_value_save.items()}
        for key,value in all_attributions_value_save.items():
            file_path = f"{save_path}{key}_integrated_gradients_value.csv"
            value.to_csv(file_path, index=True)
        for key,value in self.all_deltas.items():
            file_path = f"{save_path}{key}_approximation_error_value.csv"
            value.to_csv(file_path, index=True)
            
        return all_attributions_value_save
        
        
        
####-----不应该计算输入数据的中位数，需要先用训练集的中位数，然后再按输入数据进行填充
class DeepliftCalculator:

    def __init__(self, model, X_input, X_input_tensor,y_input,batch_size=100):
        self.model = model
        self.X_input = X_input 
        self.X_input_tensor=X_input_tensor
        self.y_input = y_input
        self.batch_size=batch_size
        
    def _compute_deepliftvalue_define(self,target_type):
        
        device = self.X_input_tensor.device  
        # 在样本维度上计算 X_input 的均值作为基线
        baseline = np.mean(self.X_input.values, axis=0,keepdims=True)

        # 使用 np.broadcast_to 将基线扩展为与 X_input 相同的形状
        baseline = np.broadcast_to(baseline, self.X_input.shape)

        # 将 numpy array 转换为 PyTorch 张量
        baseline = torch.Tensor(baseline).to(device)

        
        dl = DeepLift(self.model)

        all_attributions = []
        all_deltas = []

        num_batches = (self.X_input_tensor.shape[0] + self.batch_size - 1) // self.batch_size

        for batch_idx in range(num_batches):
            start_idx = batch_idx * self.batch_size
            end_idx = min((batch_idx + 1) * self.batch_size, self.X_input_tensor.shape[0])
            X_batch = self.X_input_tensor[start_idx:end_idx]

            attributions, delta = dl.attribute(X_batch, baseline[start_idx:end_idx], 
                            target=target_type, return_convergence_delta=True)

            attributions_np = attributions.cpu().detach().numpy()
            delta_np = delta.cpu().detach().numpy()

            all_attributions.append(attributions_np)
            all_deltas.append(delta_np)

            torch.cuda.empty_cache()
            torch.cuda.synchronize()

        all_attributions = np.concatenate(all_attributions, axis=0)
        all_deltas = np.concatenate(all_deltas, axis=0)
        
        return all_attributions, all_deltas
    
    def compute_deepliftvalue(self,type_class):
        
        self.all_attributions_values=[]
        self.all_deltas=[]
        for i in range(len(type_class)):
            all_attributions_values_inn, all_deltas_inn=self._compute_deepliftvalue_define(i)
            self.all_attributions_values.append(all_attributions_values_inn)
            self.all_deltas.append(all_deltas_inn)
        # self.all_attributions_values, self.all_deltas=list(map
        #                     (lambda x: self._compute_integrated_gradients_define(x),range(len(type_class))))
        
        self.all_attributions_values=list(map(lambda x: pd.DataFrame(x,index=self.X_input.index,columns=self.X_input.columns),self.all_attributions_values))
        self.all_deltas=list(map(lambda x: pd.DataFrame(x,index=self.X_input.index,columns=['approximation error']),self.all_deltas))
       
        self.all_attributions_values=dict(zip(type_class,self.all_attributions_values))
        self.all_deltas=dict(zip(type_class,self.all_deltas))
        
        return self.all_attributions_values, self.all_deltas
    def deepliftvalue_save(self,X_predict_input,save_path):
        all_attributions_value_save = {key: value.join(self.y_input, how="inner") for key,value in self.all_attributions_values.items()}
        all_attributions_value_save = {key: value.join(X_predict_input, how="inner") for key, value in all_attributions_value_save.items()}
        for key,value in all_attributions_value_save.items():
            file_path = f"{save_path}{key}_deeplift_value.csv"
            value.to_csv(file_path, index=True)
        for key,value in self.all_deltas.items():
            file_path = f"{save_path}{key}_approximation_error_value.csv"
            value.to_csv(file_path, index=True)
            
        return all_attributions_value_save
        
             
        
        
        
        
        
        
        
        
        
        
        
        
        


# -*- coding: utf-8 -*-
"""
Created on Fri Mar 24 18:24:37 2023

@author: qq102
"""



import pandas as pd

import re

from functools import reduce
from packaging import version
import warnings
try:
    import gensim
    if version.parse(gensim.__version__) < version.parse('4.2.0'):
        warnings.warn("Your gensim version is older than 4.2.0  +\
                      and may be not operation correct.")
except ImportError:
    raise ValueError('gensim is not installed so this module is not available! Run `pip install gensim==4.2.0` " \
                             "to fix this ')

## Building sentence processing functions
def read_corpus(sequence_input, tokens_only=False):
    '''
    
    
    Reads input text data and yields tokenized sentences or tagged sentences.
    
    Parameters
    ----------
    sequence_input : list of strings
        The input text data with per sentence.
    tokens_only : bool, optional
        If True, yields tokenized sentences only. If False, yields tagged sentences
        with sentence index as the tag. The default is False.
    Yields
    ------
    list of str or gensim.models.doc2vec.TaggedDocument
        If tokens_only is True, yields a simple preprocess for each sentence with
        no the sentence index as the tag,
        
        If tokens_only is False, yields a TaggedDocument object for each sentence
        with the sentence index as the tag. The TaggedDocument object contains
        the tokenized sentence.

    '''
    for i, line in enumerate(sequence_input):
        tokens = gensim.utils.simple_preprocess(line)
        if tokens_only:
            yield tokens
        else:
            # For training data, add tags
            yield gensim.models.doc2vec.TaggedDocument(tokens, [i])



## Building GO language string processing functions
def GO_pre_Process(GO_data,pattern_GO='\\[GO:\d+\\]',GO_name='GO_BP'):
    '''
    
    
    Pre-processes GO annotation data by removing GO ontology terms.
    
    Parameters
    ----------
    GO_data : DataFrame
        The input GO annotation data.
    pattern_GO : str, optional
        The regular expression pattern for matching GO ontology terms.
        The default is '\\[GO:\d+\\]' to match terms like [GO:0006397].
    GO_name : str, optional
        The name of the column containing GO ontology terms.
        The default is 'GO_BP'.
    Returns
    -------
    GO_data : DataFrame
        The GO annotation data with GO ontology terms (such as [GO:0006397]) 
        removed from the GO_name column.

    '''
    pattern=re.compile(pattern_GO)
    ## Removing [GO:XXXXXX] from bp, Step 1 
    for i in range(len(GO_data[GO_name])):
        GO_data.loc[i,GO_name]=re.sub(pattern,"",GO_data.loc[i,GO_name])
    return GO_data

## Building doc Embedding Generation Functions

def model_training(file_Corpus_input,model,sequence_name='GO_BP'):
    '''
    

    Trains a gensim document embedding model on input corpus data.
    
    Parameters
    ----------
    file_Corpus_input : DataFrame
        The input corpus data.
    model : gensim.models.doc2vec model
        The document embedding model to be trained.
    sequence_name : str, optional
        The name of the column containing training sentences in file_Corpus_input.
        The default is 'GO_BP'.
    Returns
    -------
    model : gensim.models.doc2vec model
        The trained document embedding model
    Note
    -------
    total_examples=model.corpus_count 
    This means we pass the value of model.corpus_count as the total_examples parameter to the model.train() function.
    The total_examples parameter specifies the total number of documents the model will train on. 
    This allows the algorithm to allocate the necessary memory in advance and improve training efficiency. 
    model.corpus_count should be the number of documents or sentences in our training corpus corpus_data. 
    So here we directly use the corpus size statistics of the model itself as the value of total_examples. 
    If total_examples is not specified, Gensim will automatically infer the corpus size, 
    but this may slightly reduce training efficiency because it has to scan the corpus twice - once to get the size statistics and once for actual training.
    In addition, the function does not support single-character sentences at now
    '''
    corpus_data=list(read_corpus(file_Corpus_input[sequence_name]))
    # Creating a vocabulary using only the corpus
    
    model.build_vocab(corpus_data)
    
    # Computing Word Embeddings from the Corpus
    model.train(corpus_data, total_examples=model.corpus_count, epochs=model.epochs)
    return model

def vec_create_GO(X_input,model,sequence_name='GO_BP',model_inference=0.5):
    '''
    

    Generates document embedding vectors for input data using the pre-trained document embedding model.
    
    Parameters
    ----------
    X_input : DataFrame
        The input data for which to generate document embeddings.
    model : gensim.models.doc2vec model
        The pre-trained document embedding model to use for generating inference embeddings.
    sequence_name : str, optional
        The name of the column in X_input containing training sentences. The default is 'GO_BP'.
    model_inference : float, optional
        The proportion of additional epochs to train the model on the new data.
        The default is 0.5.
    Raises
    ------
    ValueError
        If X_input contains duplicate index.
    Returns
    -------
    train_vector_data : DataFrame
        The document embedding vectors for X_input indexed by the same index.

    '''
    if X_input.index.duplicated().sum() > 0:
        raise ValueError("Index contains duplicates")
    ## Processing sentences in X_train and X_test
    train_corpus = list(read_corpus(X_input[sequence_name],tokens_only=True))
  
    
    # Converting data into feature embeddings

    train_vector=[]
    for doc_id in range(len(train_corpus)):
        inferred_vector=model.infer_vector(train_corpus[doc_id],
                                       epochs=int(model.epochs+model_inference*model.epochs))
        train_vector.append(inferred_vector)

  
    ## Transposing data into a pandas DataFrame

    for i,value in enumerate(train_vector):
        train_vector[i]=pd.DataFrame(value).T
    
    train_vector_data=reduce(lambda x,y:pd.concat([x,y],axis=0),train_vector)
    
 

    train_vector_data.index=X_input.index
    train_vector_data.columns=[sequence_name+str(i) for i in train_vector_data.columns]
    
    return train_vector_data



def AA_combine(X_data, sequence_name, number=3):
    '''
    

    Combines every n amino acids in a sequence into a single character.
    
    Parameters
    ----------
    X_data : DataFrame
        The input data containing an amino acid sequence column.
    sequence_name : str
        The name of the column in X_data containing the amino acid sequence.
    number : int, optional
        The number of amino acids to combine into a single character.
        The default is 3.
    Returns
    -------
    X_data : DataFrame
        The input data with the amino acid sequence column modified to
        contain combined amino acid characters.


    '''
    X_data = X_data.copy()

    def combine_sequence(value):
        # Initializing an empty sequence string 
        sequence_inner = ""
        # Iterating over an amino acid sequence
        
        for i in range(0, len(value), number):
            # Extracting 3 amino acids
            triplet = value[i:i+number]
            # Concatenating 3 amino acids into one character and adding spaces
            sequence_inner += triplet + " "

        return sequence_inner

    X_data[sequence_name] = X_data[sequence_name].apply(combine_sequence)

    return X_data


# -*- coding: utf-8 -*-
"""
Created on Fri Mar 10 14:43:58 2023

@author: qq102
"""
import shap
import logging
import matplotlib.pyplot as plt
import pandas as pd
import numpy as np


class SHAPValueConduct:
    '''
    A class used to analyze SHAP values.
    
    Attributes:
    ----------
    explainer : SHAP explainer 
        The explainer used to calculate SHAP values, 
        for now are only support shap.TreeExplainer.
    X_input : DataFrame
        Input features.
    y_input : Series, DataFrame
        Target values.
    
    '''
    def __init__(self, explainer,X_input,y_input):
        self.explainer = explainer
        self.X_input=X_input
        self.y_input=y_input
    def shap_value_conduct(self): 
        '''
        

        Calculate the SHAP values of X_input using explainer.

        Returns
        -------
        shap_values : list
            SHAP values for each sample in X_input for each localization.

        '''
        self.shap_values = self.explainer.shap_values(self.X_input)
        return self.shap_values
    
    def save_shap_values(self, X_predict_input,shap_value_input,X_input,y_input, type_class,save_path,file_name,gene_ID):
        '''
        

        Save SHAP values, input features, target values and predictions to CSV file.  

        Parameters
        ----------
        X_predict_input : DataFrame
            Model predictions for X_input. 
        shap_value_input : list
            SHAP values for each sample in X_input for One localization.
        X_input : DataFrame
            Input features.
        y_input : Series, DataFrame
            Target values.
        type_class : str
            One target class name.               
        save_path : str
            input is file or plot save path way with no contain file name.
        file_name : str
            only indicate test or train but not specific file name, 
            such as input can't be classification_report.csv.
        gene_ID : str
            Name of the index column.

        Returns
        -------
        X_shap_save : DataFrame
            Combined information of SHAP values, inputs, targets and predictions.

        '''
                
        X_shap_save = pd.DataFrame(shap_value_input, columns=X_input.columns, index=X_input.index)
        X_shap_save = X_shap_save.join(y_input, how="inner")
        X_shap_save = X_shap_save.join(X_predict_input, how="inner")
        file_path = f"{save_path}/{file_name}_{type_class}_shap_value.csv"
        X_shap_save.to_csv(file_path, index_label=gene_ID)
        return X_shap_save
        

    def Shapley_value_save(self,X_predict_input,type_class,save_path,file_name="train",gene_ID='ID'):
        '''
        

        Save SHAP values for different classes.
        
        Parameters
        ----------
        X_predict_input : DataFrame
            Model predictions for X_input.  
        type_class : list
            Names of the target classes.                 
        save_path : str
            input is file or plot save path way with no contain file name.
        file_name : str, optional
            only indicate test or train but not specific file name, The default is "train".
        gene_ID : str, optional
            Name of the index column. The default is 'ID'.
        
        Returns
        -------
        X_shap_save : dict
            Saved information of SHAP values (SHAP values, inputs, targets and predictions) for each class.
        
        Note
        -------
        Please ensure that the type entered for the type_class parameter matches 
        the type of the model default classification,
        otherwise the results may be inaccurate!
        '''
        logging.warning('Please ensure that the type entered for the type_class parameter matches the type of the model default classification,\
                        otherwise the results may be inaccurate!')
        X_shap_save=list(map(lambda x,y:self.save_shap_values(X_predict_input,\
                                    self.shap_values[y],self.X_input,self.y_input,x,save_path,file_name,gene_ID),type_class,range(len(type_class))))
        X_shap_save=dict(zip(type_class,X_shap_save))
        return X_shap_save


class ShapInteractionIndexConduct:
    '''
    
    
    A class used to analyze SHAP interaction values.
    
    Attributes:
    ----------
    frac_num : float
        The fraction of samples (random select from original samples) used to calculate SHAP interaction values.
    tree_limit_num : int
        The number of trees used to calculate SHAP interaction values, suggest setting is 100.
    explainer : tree explainer
        The explainer used to calculate SHAP interaction values,
        for now are only support shap.TreeExplainer..
    shap_plot_figure_size : tuple
        The figure size of the SHAP summary interaction plot.
    
    '''
    
    def __init__(self, frac_num, tree_limit_num, explainer, shap_plot_figure_size):
        self.frac_num = frac_num
        self.tree_limit_num = tree_limit_num
        self.explainer = explainer
        self.shap_plot_figure_size = shap_plot_figure_size
    
    def plot_shap_interaction_index_values(self, shap_interaction_values, X_input, class_names, file_name, save_path):
        '''
        

        Plot the SHAP summary interaction plot and save the figure.
        
        Parameters:
        -----------
        shap_interaction_values : list
            SHAP interaction values for each sample in X_input for one localization.
        X_input : DataFrame
            input features from randomly selected dataframe.          
        class_names : str
            Name of the target class.
        file_name : str
            only indicate test or train but not specific file name, 
            such as input can't be classification_report.csv.
        save_path : str
            input is file or plot save path way with no contain file name.
            
            
        '''
        
        shap.summary_plot(shap_interaction_values, X_input, 
                           plot_size=self.shap_plot_figure_size, show=False, max_display=10)
        
        
        plt.savefig(save_path + file_name + "_" + class_names  + "_shap_interaction.png",
                    dpi=1000, bbox_inches="tight")
        plt.savefig(save_path + file_name + "_" + class_names  + "_shap_interaction.pdf",
                    dpi=1000, bbox_inches="tight")
        plt.close()

    def save_shap_interaction_index(self, shap_interaction_values, X_random, file_name, save_path,type_class):
        '''
        

        Save the SHAP interaction values for different classes.
        
        Parameters:
        -----------
        shap_interaction_values : list
            SHAP interaction values for each sample in X_random. 
        X_random : DataFrame
             input features from randomly selected dataframe.          
        file_name : str
            only indicate test or train but not specific file name, 
            such as input can't be classification_report.csv.
        save_path : str
            input is file or plot save path way with no contain file name.
        type_class : list
            Names of the target classes.
       
      
        '''
        for i, value in enumerate(type_class):
            self.plot_shap_interaction_index_values(shap_interaction_values[i], X_random, value,  file_name, save_path)

            shap_interaction_data = pd.DataFrame(shap_interaction_values[i][0,:,:],
                                                       index=X_random.columns, columns=X_random.columns)
            shap_interaction_data['feature_max'] = shap_interaction_data.apply(lambda x: sorted(abs(x))[-2], axis='columns')
            shap_interaction_data.to_csv(save_path + file_name + "_" + value + "_dim0_shap_interaction.csv", index_label='feature_name')

            shap_interaction_abs = pd.DataFrame(np.abs(shap_interaction_values[i]).sum(0),
                                                      index=X_random.columns, columns=X_random.columns)
            shap_interaction_abs.to_csv(save_path  + file_name + "_" + value + "_shap_interaction_abs.csv", index_label='feature_name')

        shap_interaction_values_all = dict(zip(type_class, shap_interaction_values))
        X_random.to_csv(save_path + file_name + "_feature_random.csv", index_label='ID')
        np.save(save_path + file_name + "_shap_interaction_all.npy", shap_interaction_values_all)

    def conduct(self, X_input, file_name, save_path, type_class):
        '''
        

        Conduct the analysis of SHAP interaction values.
        
        Parameters:
        -----------
        X_input : DataFrame        
             input features from full dataframe.          
        file_name : str
            only indicate test or train but not specific file name, 
            such as input can't be classification_report.csv.
        save_path : str
            input is file or plot save path way with no contain file name.
        type_class : list
            Names of the target classes.  
            
        Output:
        --------
        SHAP summary interaction plot (PNG and PDF format)
        SHAP interaction strength data (CSV format), Note: calculate format are displayed in <under review article>   
        Features data randomly selected from original dataframe (CSV format)
        SHAP interaction values for all classes (NPY format)
        
        Note
        -------
        Please ensure that the type entered for the type_class parameter matches 
        the type of the model default classification,
        otherwise the results may be inaccurate!
        '''
        logging.warning('Please ensure that the type entered for the type_class parameter matches the type of the model default classification,\
                        otherwise the results may be inaccurate!')

        X_random = X_input.sample(frac=self.frac_num, replace=False, random_state=0, axis=0)
        shap_interaction_values = self.explainer.shap_interaction_values(X_random, tree_limit=self.tree_limit_num)

        self.save_shap_interaction_index(shap_interaction_values, X_random, file_name, save_path,type_class)



class DeepSHAPValueConduct:
    '''
    A class used to analyze Deep SHAP values.
    
    Attributes:
    ----------
    explainer : SHAP DeepExplainer 
        The explainer used to calculate SHAP values.
    X_input : DataFrame
        Input features.
    y_input : Series, DataFrame
        Target values.
    '''
    def __init__(self, explainer , X_input, X_input_tensor,y_input):
        
        
        # Randomly select samples from X_input to form the background dataset
        
        self.explainer = explainer 
        self.X_input = X_input
        self.X_input_tensor=X_input_tensor
        self.y_input = y_input
    
    def shap_value_conduct(self):
        '''
        Calculate the SHAP values of X_input using explainer.

        Returns
        -------
        shap_values : list
            SHAP values for each sample in X_input for each localization.
        '''
        self.shap_values = self.explainer.shap_values(self.X_input_tensor)  # Convert DataFrame to numpy array
        return self.shap_values

    def save_shap_values(self, X_predict_input,shap_value_input,X_input,y_input, type_class,save_path,file_name,gene_ID):
        '''
        

        Save SHAP values, input features, target values and predictions to CSV file.  

        Parameters
        ----------
        X_predict_input : DataFrame
            Model predictions for X_input. 
        shap_value_input : list
            SHAP values for each sample in X_input for One localization.
        X_input : DataFrame
            Input features.
        y_input : Series, DataFrame
            Target values.
        type_class : str
            One target class name.               
        save_path : str
            input is file or plot save path way with no contain file name.
        file_name : str
            only indicate test or train but not specific file name, 
            such as input can't be classification_report.csv.
        gene_ID : str
            Name of the index column.

        Returns
        -------
        X_shap_save : DataFrame
            Combined information of SHAP values, inputs, targets and predictions.

        '''
                
        X_shap_save = pd.DataFrame(shap_value_input, columns=X_input.columns, index=X_input.index)
        X_shap_save = X_shap_save.join(y_input, how="inner")
        X_shap_save = X_shap_save.join(X_predict_input, how="inner")
        file_path = f"{save_path}/{file_name}_{type_class}_shap_value.csv"
        X_shap_save.to_csv(file_path, index_label=gene_ID)
        return X_shap_save
        

    def Shapley_value_save(self,X_predict_input,type_class,save_path,file_name="train",gene_ID='ID'):
        '''
        

        Save SHAP values for different classes.
        
        Parameters
        ----------
        X_predict_input : DataFrame
            Model predictions for X_input.  
        type_class : list
            Names of the target classes.                 
        save_path : str
            input is file or plot save path way with no contain file name.
        file_name : str, optional
            only indicate test or train but not specific file name, The default is "train".
        gene_ID : str, optional
            Name of the index column. The default is 'ID'.
        
        Returns
        -------
        X_shap_save : dict
            Saved information of SHAP values (SHAP values, inputs, targets and predictions) for each class.
        
        Note
        -------
        Please ensure that the type entered for the type_class parameter matches 
        the type of the model default classification,
        otherwise the results may be inaccurate!
        '''
        logging.warning('Please ensure that the type entered for the type_class parameter matches the type of the model default classification,\
                        otherwise the results may be inaccurate!')
        X_shap_save=list(map(lambda x,y:self.save_shap_values(X_predict_input,\
                                    self.shap_values[y],self.X_input,self.y_input,x,save_path,file_name,gene_ID),type_class,range(len(type_class))))
        X_shap_save=dict(zip(type_class,X_shap_save))
        return X_shap_save
    
    
    









from Bio.SeqUtils.ProtParam import ProteinAnalysis

import numpy as np

import re

import pandas as pd

def dna_sequence_conduct(df,sequence_name= 'gene_seq'):
    
    elements = {'C': {'A': 5, 'T': 5, 'C': 4, 'G': 5},
                'H': {'A': 5, 'T': 6, 'C': 5, 'G': 5},
                'N': {'A': 5, 'T': 2, 'C': 3, 'G': 5},
                'O': {'A': 0, 'T': 2, 'C': 1, 'G': 1}}

    # 过滤所有非法碱基
    df[sequence_name] = df[sequence_name].apply(lambda x: ''.join([base for base in x if base in {'A', 'T', 'C', 'G'}]))
    # 计算元素的频率
    df['Carbon']   = df[sequence_name].apply(lambda x: sum([elements['C'][base] for base in x]) / len(x))
    df['Hydrogen'] = df[sequence_name].apply(lambda x: sum([elements['H'][base] for base in x]) / len(x))
    df['Nitrogen'] = df[sequence_name].apply(lambda x: sum([elements['N'][base] for base in x]) / len(x))
    df['Oxygen']   = df[sequence_name].apply(lambda x: sum([elements['O'][base] for base in x]) / len(x))
    # 计算碱基频率
    df['A'] = df[sequence_name].apply(lambda x: x.count('A') / len(x))
    df['T'] = df[sequence_name].apply(lambda x: x.count('T') / len(x))
    df['G'] = df[sequence_name].apply(lambda x: x.count('G') / len(x))
    df['C'] = df[sequence_name].apply(lambda x: x.count('C') / len(x))
    
    
    return df

def rna_sequence_conduct(df, sequence_name= 'gene_seq'):
    
    elements = {'C': {'A': 5, 'U': 4, 'C': 4, 'G': 5},
                'H': {'A': 5, 'U': 4, 'C': 5, 'G': 5},
                'N': {'A': 5, 'U': 2, 'C': 3, 'G': 5},
                'O': {'A': 0, 'U': 2, 'C': 1, 'G': 1}}

    # 过滤所有非法碱基
    df[sequence_name] = df[sequence_name].apply(lambda x: ''.join([base for base in x if base in {'A', 'U', 'C', 'G'}]))
    # 计算元素的频率
    df['Carbon']   = df[sequence_name].apply(lambda x: sum([elements['C'][base] for base in x]) / len(x))
    df['Hydrogen'] = df[sequence_name].apply(lambda x: sum([elements['H'][base] for base in x]) / len(x))
    df['Nitrogen'] = df[sequence_name].apply(lambda x: sum([elements['N'][base] for base in x]) / len(x))
    df['Oxygen']   = df[sequence_name].apply(lambda x: sum([elements['O'][base] for base in x]) / len(x))
    # 计算碱基频率
    df['A'] = df[sequence_name].apply(lambda x: x.count('A') / len(x))
    df['U'] = df[sequence_name].apply(lambda x: x.count('U') / len(x))
    df['G'] = df[sequence_name].apply(lambda x: x.count('G') / len(x))
    df['C'] = df[sequence_name].apply(lambda x: x.count('C') / len(x))
    
    
    return df

def protein_sequence_conduct(file_data, sequence_name="Sequence"):
    """
    Conduct protein sequence analysis and add relevant protein properties to an input pandas dataframe.

    Parameters
    ----------
        file_data : pd.DataFrame 
            Input pandas dataframe containing protein sequences.
        sequence_name : str 
            Column name of the protein sequences in the input dataframe (default is "Sequence").

    Returns:
    ----------
        pd.DataFrame: Output pandas dataframe with added protein properties columns.

    Raises:
    ----------
        ValueError: If the index of the input dataframe contains duplicates.

    Example:
    ----------
        >>> import pandas as pd
        >>> from protloc_mex1.AA_count import protein_sequence_conduct
        >>> df = pd.DataFrame({'Sequence': ['KKSAAEKKKKKKKKAH', 'AKDFLIEAELKDSFF']})
        >>> output_df = protein_sequence_conduct(df,sequence_name="Sequence")

    Notes:
    ----------
        This function utilizes the ProteinAnalysis function from the Bio.SeqUtils.ProtParam module in the Biopython
        package. The input dataframe should contain a column with protein sequences. The function will add columns with
        protein properties including amino acid frequencies, elemental properties, molecular weight, aromaticity,
        instability index, flexibility, isoelectric point, secondary structure fraction, molar extinction coefficient
        for reduced cysteines and disulfid bridges, gravy, and a boolean column indicating if the sequence contains
        only standard amino acids.
    """

    # Define a function to check if the sequence contains rare amino acids or unknown amino acids.
    # O stands for pyrrolysine, Pyl; U stands for selenocysteine, Sec.
    # def special_match(strg):
    #     pattern = re.compile(r'[U|X|O|Z]+')
    #     return not bool(pattern.findall(strg))
    
    if file_data.index.duplicated().sum() > 0:
        raise ValueError("Index contains duplicates")

    # Create columns for the 20 amino acids
    AA_columns = ['A','C','D','E','F','G','H','I','K','L','M','N','P','Q','R','S','T','V','W','Y']
    # Create columns for elemental properties
    element_count = ["Oxygen","Sulfur","Carbon","Hydrogen","Nitrogen","Acid","Basic","Polar","Non_Polar"]
    # Create columns for protein properties
    protein_features = AA_columns + element_count + ["molecular_weight","aromaticity","instability_index",
                                                    "flexibility_mean","flexibility_max","flexibility_min",
                                                    "isoelectric_point"] + \
                        ["gravy", #"protein_scale",
                         "secondary_structure_fraction_a","secondary_structure_fraction_b1",
                         "secondary_structure_fraction_b2","molar_extinction_coefficient_reduced_cysteines",
                         "molar_extinction_coefficient_disulfid_bridges"] + ["normal_sequence"]
    
    # Define a function to determine if it contains only 20 standard amino acids and no other characters
    def special_match(strg):
        pattern = re.compile(r'[^' + ''.join(AA_columns) + ']')
        return not bool(pattern.search(strg))

    # Get the column names to be added
    new_columns = list(set(protein_features) - set(file_data.columns))
    new_columns = [col for col in protein_features if col in new_columns]

    # Add feature columns based on added column name
    if new_columns:
        file_data = file_data.reindex(columns=list(file_data.columns) + new_columns)

    
    # Determine if the sequence contains rare amino acids or unknown amino acids
    for i in file_data.index:
        file_data.loc[i, "normal_sequence"] = special_match(file_data.loc[i, sequence_name])
    
    

      
    for i in file_data.index:
        Seq_count = ProteinAnalysis(file_data.loc[i, sequence_name])
        # calculate frequency of 20 amino acids
        AA_conduct_output = Seq_count.get_amino_acids_percent()

        for AA_conduct in AA_columns:
            file_data.loc[i, AA_conduct] = AA_conduct_output[AA_conduct]
        if file_data.loc[i, "normal_sequence"]:
            # calculate molecular weight
            file_data.loc[i, "molecular_weight"] = Seq_count.molecular_weight()
            # calculate aromaticity
            file_data.loc[i, "aromaticity"] = Seq_count.aromaticity()
            # calculate isoelectric point
            file_data.loc[i, "isoelectric_point"] = Seq_count.isoelectric_point()
            # calculate secondary structure fraction
            file_data.loc[i, "secondary_structure_fraction_a"] = Seq_count.secondary_structure_fraction()[0]
            file_data.loc[i, "secondary_structure_fraction_b1"] = Seq_count.secondary_structure_fraction()[1]
            file_data.loc[i, "secondary_structure_fraction_b2"] = Seq_count.secondary_structure_fraction()[2]
            # calculate molar extinction coefficient for reduced cysteines and disulfid bridges
            file_data.loc[i, "molar_extinction_coefficient_reduced_cysteines"] = Seq_count.molar_extinction_coefficient()[0]
            file_data.loc[i, "molar_extinction_coefficient_disulfid_bridges"] = Seq_count.molar_extinction_coefficient()[1]
            # calculate instability index
            file_data.loc[i, "instability_index"] = Seq_count.instability_index()
            # calculate flexibility (mean, min, max)
            flexibility_vals = Seq_count.flexibility()
            if len(flexibility_vals) > 0:
                file_data.loc[i, "flexibility_mean"] = np.mean(flexibility_vals)
                file_data.loc[i, "flexibility_min"] = np.min(flexibility_vals)
                file_data.loc[i, "flexibility_max"] = np.max(flexibility_vals)
            else:
                file_data.loc[i, "flexibility_mean"] = np.nan
                file_data.loc[i, "flexibility_min"] = np.nan
                file_data.loc[i, "flexibility_max"] = np.nan
            
            # calculate gravy
            file_data.loc[i, "gravy"] = Seq_count.gravy(scale='KyteDoolitle')
            # calculate protein scale
            # file_data.loc[i, "protein_scale"] = Seq_count.protein_scale(edge=1.0,)



        

    ## Calculate amino acid elemental composition
    for i in file_data.index:
        ##计算Oxygen
        file_data.loc[i, "Oxygen"]=(file_data.loc[i, sequence_name].count("Y")+file_data.loc[i, sequence_name].count("S")+\
        file_data.loc[i, sequence_name].count("T")+file_data.loc[i, sequence_name].count("N")+\
        file_data.loc[i, sequence_name].count("Q")+2*file_data.loc[i, sequence_name].count("D")+\
        2*file_data.loc[i, sequence_name].count("E"))/len(file_data.loc[i, sequence_name])
        ##Calculate Sulfur
        file_data.loc[i,"Sulfur"]=(file_data.loc[i, sequence_name].count("M")+file_data.loc[i, sequence_name].count("C"))/len(file_data.loc[i, sequence_name])
        ##Calculate Carbon
        file_data.loc[i,"Carbon"]=(file_data.loc[i, sequence_name].count("A")+\
                              3*file_data.loc[i, sequence_name].count("V")+4*file_data.loc[i, sequence_name].count("L")+\
                              4*file_data.loc[i, sequence_name].count("I")+3*file_data.loc[i, sequence_name].count("M")+\
                              7*file_data.loc[i, sequence_name].count("F")+7*file_data.loc[i, sequence_name].count("Y")+\
                              9*file_data.loc[i, sequence_name].count("W")+\
                              file_data.loc[i, sequence_name].count("S")+3*file_data.loc[i, sequence_name].count("P")+\
                              2*file_data.loc[i, sequence_name].count("T")+file_data.loc[i, sequence_name].count("C")+\
                              2*file_data.loc[i, sequence_name].count("N")+3*file_data.loc[i, sequence_name].count("Q")+\
                              4*file_data.loc[i, sequence_name].count("K")+4*file_data.loc[i, sequence_name].count("H")+\
                              4*file_data.loc[i, sequence_name].count("R")+2*file_data.loc[i, sequence_name].count("D")+\
                              3*file_data.loc[i, sequence_name].count("E"))/len(file_data.loc[i, sequence_name])
        ##Calculate Hydrogen
        file_data.loc[i,"Hydrogen"]=(file_data.loc[i, sequence_name].count("G")+3*file_data.loc[i, sequence_name].count("A")+\
                              7*file_data.loc[i, sequence_name].count("V")+9*file_data.loc[i, sequence_name].count("L")+\
                              9*file_data.loc[i, sequence_name].count("I")+7*file_data.loc[i, sequence_name].count("M")+\
                              7*file_data.loc[i, sequence_name].count("F")+7*file_data.loc[i, sequence_name].count("Y")+\
                              8*file_data.loc[i, sequence_name].count("W")+\
                              3*file_data.loc[i, sequence_name].count("S")+6*file_data.loc[i, sequence_name].count("P")+\
                              5*file_data.loc[i, sequence_name].count("T")+3*file_data.loc[i, sequence_name].count("C")+\
                              4*file_data.loc[i, sequence_name].count("N")+6*file_data.loc[i, sequence_name].count("Q")+\
                              11*file_data.loc[i, sequence_name].count("K")+6*file_data.loc[i, sequence_name].count("H")+\
                              11*file_data.loc[i, sequence_name].count("R")+2*file_data.loc[i, sequence_name].count("D")+\
                              4*file_data.loc[i, sequence_name].count("E"))/len(file_data.loc[i, sequence_name])
        ##Calculate Nitrogen
        file_data.loc[i,"Nitrogen"]=(file_data.loc[i, sequence_name].count("K")+3*file_data.loc[i, sequence_name].count("R")+\
        2*file_data.loc[i, sequence_name].count("H")+file_data.loc[i, sequence_name].count("W")+\
        file_data.loc[i, sequence_name].count("N")+file_data.loc[i, sequence_name].count("Q"))/len(file_data.loc[i, sequence_name])
        ##Calculate Acid
        file_data.loc[i,"Acid"]=(file_data.loc[i, sequence_name].count("D")+file_data.loc[i, sequence_name].count("E"))/len(file_data.loc[i, sequence_name])
        ##Calculate Basic
        file_data.loc[i,"Basic"]=(file_data.loc[i, sequence_name].count("H")+file_data.loc[i, sequence_name].count("R")+\
                                file_data.loc[i, sequence_name].count("K"))/len(file_data.loc[i, sequence_name])
        ##Calculate Polar
        file_data.loc[i,"Polar"]=(file_data.loc[i, sequence_name].count("S")+file_data.loc[i, sequence_name].count("P")+\
                              file_data.loc[i, sequence_name].count("T")+file_data.loc[i, sequence_name].count("C")+\
                              file_data.loc[i, sequence_name].count("N")+file_data.loc[i, sequence_name].count("Q")+\
                              file_data.loc[i, sequence_name].count("K")+file_data.loc[i, sequence_name].count("H")+\
                              file_data.loc[i, sequence_name].count("R")+file_data.loc[i, sequence_name].count("D")+\
                              file_data.loc[i, sequence_name].count("E"))/len(file_data.loc[i, sequence_name])
        ##Calculate Non_Polar
        file_data.loc[i,"Non_Polar"]=(file_data.loc[i, sequence_name].count("G")+file_data.loc[i, sequence_name].count("A")+\
                              file_data.loc[i, sequence_name].count("V")+file_data.loc[i, sequence_name].count("L")+\
                              file_data.loc[i, sequence_name].count("I")+file_data.loc[i, sequence_name].count("M")+\
                              file_data.loc[i, sequence_name].count("F")+file_data.loc[i, sequence_name].count("Y")+\
                              file_data.loc[i, sequence_name].count("W"))/len(file_data.loc[i, sequence_name])
    return file_data
 

def process_protein_data(data, Length='Length', normal_sequence='normal_sequence'):
    '''

    Parameters
    ----------
    data : pd.Dataframe
        output of AA_count.protein_sequence_conduct.
    Length : TYPE, optional
        DESCRIPTION. The default is 'Length'.
    normal_sequence : TYPE, optional
        DESCRIPTION. The default is 'normal_sequence'.

    Returns
    -------
    TYPE
        DESCRIPTION.

    '''
    # 将'normal_sequence'列转换为字符串类型
    data[normal_sequence] = data[normal_sequence].astype(str)

    if data[normal_sequence].str.contains('False').any():
        # 提取含有False的行保存为minor_base_data
        minor_base_data = data[data[normal_sequence] == 'False']
        # 提取不含False的行保存为normal_base_data
        normal_base_data = data[data[normal_sequence] != 'False']

    else:
        # 如果不含有False，则直接输出提示信息
        print("The dataset does not exhibit the presence of uncommon nucleotides.")
        # return None

    # 如果normal_base_data不为空
    if not minor_base_data.empty:
        # 依次处理minor_base_data中的每一行
        for _, row in minor_base_data.iterrows():
            target_length = row[Length]
            # 在normal_base_data中找出与目标长度相差在10%以内的行
            similar_rows = normal_base_data[
                (normal_base_data[Length] >= target_length * 0.9) &
                (normal_base_data[Length] <= target_length * 1.1)
            ].sort_values(by=Length, key=lambda x: abs(x - target_length))

            zero_rows = similar_rows[similar_rows[Length] == target_length]

            if len(zero_rows) >= 5:
                similar_rows = zero_rows
            else:
                closest_rows = similar_rows.head(5)
                similar_rows = closest_rows

            if len(similar_rows) >= 5:
                # 计算每列的平均值
                # 将平均值赋给minor_base_data对应列
                minor_base_data.at[_, 'molecular_weight'] = similar_rows['molecular_weight'].mean()
                minor_base_data.at[_, 'aromaticity'] = similar_rows['aromaticity'].mean()
                minor_base_data.at[_, 'instability_index'] = similar_rows['instability_index'].mean()
                minor_base_data.at[_, 'flexibility_mean'] = similar_rows['flexibility_mean'].mean()
                minor_base_data.at[_, 'flexibility_max'] = similar_rows['flexibility_max'].mean()
                minor_base_data.at[_, 'flexibility_min'] = similar_rows['flexibility_min'].mean()
                minor_base_data.at[_, 'isoelectric_point'] = similar_rows['isoelectric_point'].mean()
                minor_base_data.at[_, 'gravy'] = similar_rows['gravy'].mean()
                minor_base_data.at[_, 'secondary_structure_fraction_a'] = similar_rows['secondary_structure_fraction_a'].mean()
                minor_base_data.at[_, 'secondary_structure_fraction_b1'] = similar_rows['secondary_structure_fraction_b1'].mean()
                minor_base_data.at[_, 'secondary_structure_fraction_b2'] = similar_rows['secondary_structure_fraction_b2'].mean()
                minor_base_data.at[_, 'molar_extinction_coefficient_reduced_cysteines'] = similar_rows['molar_extinction_coefficient_reduced_cysteines'].mean()
                minor_base_data.at[_, 'molar_extinction_coefficient_disulfid_bridges'] = similar_rows['molar_extinction_coefficient_disulfid_bridges'].mean()
            else:
                # 如果相差在10%以内的数量少于5，则计算为整个
                minor_base_data.at[_, 'molecular_weight'] = normal_base_data['molecular_weight'].mean()
                minor_base_data.at[_, 'aromaticity'] = normal_base_data['aromaticity'].mean()
                minor_base_data.at[_, 'instability_index'] = normal_base_data['instability_index'].mean()
                minor_base_data.at[_, 'flexibility_mean'] = normal_base_data['flexibility_mean'].mean()
                minor_base_data.at[_, 'flexibility_max'] = normal_base_data['flexibility_max'].mean()
                minor_base_data.at[_, 'flexibility_min'] = normal_base_data['flexibility_min'].mean()
                minor_base_data.at[_, 'isoelectric_point'] = normal_base_data['isoelectric_point'].mean()
                minor_base_data.at[_, 'gravy'] = normal_base_data['gravy'].mean()
                minor_base_data.at[_, 'secondary_structure_fraction_a'] = normal_base_data['secondary_structure_fraction_a'].mean()
                minor_base_data.at[_, 'secondary_structure_fraction_b1'] = normal_base_data['secondary_structure_fraction_b1'].mean()
                minor_base_data.at[_, 'secondary_structure_fraction_b2'] = normal_base_data['secondary_structure_fraction_b2'].mean()
                minor_base_data.at[_, 'molar_extinction_coefficient_reduced_cysteines'] = normal_base_data['molar_extinction_coefficient_reduced_cysteines'].mean()
                minor_base_data.at[_, 'molar_extinction_coefficient_disulfid_bridges'] = normal_base_data['molar_extinction_coefficient_disulfid_bridges'].mean()

    # 将两个结果按行合并起来
        df_AAcount_data = pd.concat([minor_base_data, normal_base_data], ignore_index=True)

        return df_AAcount_data
    else:
        return data      




import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from functools import reduce
import logging

from packaging import version
import warnings
import seaborn as sns
try:
    import sklearn
    if version.parse(sklearn.__version__) < version.parse('1.0.2'):
        warnings.warn("Your sklearn version is older than 1.0.2 and may not operate correctly.")
    from sklearn.metrics import roc_curve, auc, confusion_matrix
    from sklearn.metrics import classification_report, accuracy_score, roc_auc_score
    from sklearn.metrics import matthews_corrcoef
except ImportError:
    warnings.warn("Sklearn not found. Some functions will not be available.")

# try:
#     import mglearn
# except ImportError:
#     warnings.warn("Mglearn not found. Some functions will not be available.")

    
class ClassifierEvaluator:
    """
    A class for evaluating the performance of a classifier model.
    
    Attributes:
    ----------
        model_prob  : np.array
            input is model.predict_proba(X_data)
        y_data : np.array or pd.Series, pd.DataFrame
            contain the true labels of the samples.
        X_hat_data : pd.DataFrame
            A Pandas DataFrame containing the predicted class labels for each sample with columns is predict.
        type_class : np.array
            A array of the class labels and must as the same order as type of the model default classification.
        
    """
    def __init__(self, model_prob, y_data,X_hat_data,type_class):
        self.model_prob = model_prob
        self.y_data = np.array(y_data)
        self.X_hat_data=X_hat_data
        self.type_class=type_class
        logging.warning('Please ensure that the type entered for the type_class parameter matches the type of the model default classification,\
                        otherwise the results may be inaccurate!')
    class ClassificationReportConduct:
        def __init__(self, func):
            self.func = func
            self.self_obj = None
        
        def __get__(self, instance, owner):
            self.self_obj = instance
            return self
        
        def __call__(self, *args, **kwargs):
            self.plot(*args, **kwargs)
        
        def accuracy_calculate(self,protein_class):
            
            type_mapping={}
            for value in self.self_obj.type_class:
                if value!=protein_class:
                    type_mapping[value]='other'
                else:
                    type_mapping[value]=value
            y_true=pd.Series(self.self_obj.y_data).map(type_mapping)  #Sample actual values
            y_pred=self.self_obj.X_hat_data["predict"].map(type_mapping) #Model Predicted Values
            accuracy=accuracy_score(y_true,y_pred)
            accuracy_save_data=pd.DataFrame({protein_class:["{:.5f}".format(accuracy)]},columns=[protein_class]
                                            ,index=['accuracy'])
            return(accuracy_save_data)
        
        def auc_calculate(self,protein_class):
           
            type_mapping={}
            for value in self.self_obj.type_class:
                if value!=protein_class:
                    type_mapping[value]=0
                else:
                    type_mapping[value]=1
            y_true=pd.Series(self.self_obj.y_data).map(type_mapping)  #Sample actual values
            y_pred=self.self_obj.X_hat_data["predict"].map(type_mapping) #Model Predicted Values
            auc=roc_auc_score(y_true,y_pred)
            auc_save_data=pd.DataFrame({protein_class:["{:.5f}".format(auc)]},columns=[protein_class]
                                            ,index=['auc'])
            return(auc_save_data)
        
        def mcc_calculate(self, protein_class):
            type_mapping = {}
            for value in self.self_obj.type_class:
                if value != protein_class:
                    type_mapping[value] = 0
                else:
                    type_mapping[value] = 1
            y_true = pd.Series(self.self_obj.y_data).map(type_mapping)  # Sample actual values
            y_pred = self.self_obj.X_hat_data["predict"].map(type_mapping) # Model Predicted Values
            mcc = matthews_corrcoef(y_true, y_pred)
            mcc_save_data = pd.DataFrame({protein_class: ["{:.5f}".format(mcc)]}, columns=[protein_class], index=['mcc'])
            return mcc_save_data
    
        
        def plot(self,save_path,file_name):
            
            ##accuracy
            accuracy_save_list=list(map(lambda x:self.accuracy_calculate(x),self.self_obj.type_class))
            accuracy_save_data=reduce(lambda x,y:pd.concat([x,y],axis=1),accuracy_save_list)
            ##auc
            auc_save_list=list(map(lambda x:self.auc_calculate(x),self.self_obj.type_class))
            auc_save_data=reduce(lambda x,y:pd.concat([x,y],axis=1),auc_save_list)
            ##mcc
            mcc_save_list = list(map(lambda x: self.mcc_calculate(x), self.self_obj.type_class))
            mcc_save_data = reduce(lambda x, y: pd.concat([x, y], axis=1), mcc_save_list)

            ## Merging and save accuracy and auc
            accuracy_auc_mcc_outcome=pd.concat([accuracy_save_data,auc_save_data,mcc_save_data],axis=0)
            accuracy_auc_mcc_outcome.to_csv(save_path+file_name+'_classification_report2.csv',index_label='index')
            ## Save f1, recall, precision
            test_class_report=classification_report(self.self_obj.y_data, 
                                                    self.self_obj.X_hat_data["predict"].to_numpy(),
                                                    output_dict=True)
            ## Save test set results
            test_class_report_save=dict()
            for i in self.self_obj.type_class.tolist():
                test_class_report_save[i]=test_class_report.get(i)
            for i in test_class_report_save.keys():
                test_class_report_save[i]=pd.Series(test_class_report_save[i],name=i)
            test_class_report_save=pd.DataFrame(reduce(lambda x,y:pd.concat([x,y],axis=1),test_class_report_save.values()))
            test_class_report_save.to_csv(save_path+file_name+'_classification_report.csv',index_label='index')
    @ClassificationReportConduct
    def classification_report_conduct(self,save_path,file_name, *args, **kwargs):
        """
        Parameters
        ----------
        save_path : str
            input is file or plot save path way with no contain file name.
        file_name : str
            only indicate test or train but not specific file name, 
            such as input can't be classification_report.csv.
        *args : NoneType
            Placeholder parameter, no real meaning for now.
        **kwargs : NoneType
            Placeholder parameter, no real meaning for now.

        Returns
        -------
        Export the files of evaluation metrics such as f1-score of the classification 
        to the specified path.
        output : _classification_report.csv and _classification_report2.csv
        
        """
        pass

            


    class ClassificationEvaluatePlot:
        def __init__(self, func):
            self.func = func
            self.self_obj = None
        
        def __get__(self, instance, owner):
            self.self_obj = instance
            return self
        
        def __call__(self, *args, **kwargs):
            self.plot(*args, **kwargs)
        
        def two_features_roc(self):
            
            type_mapping={}
            type_mapping = {self.self_obj.type_class[0]: 0,
                                 self.self_obj.type_class[1]: 1}
            y_true = pd.Series(self.self_obj.y_data).map(type_mapping)
            # calculate target probabilities
            target_probabilities = self.self_obj.model_prob[:,1]
            # compute false positive rate, true positive rate, and threshold
            false_positive_rate, true_posistive_rate, threshold = roc_curve(y_true, target_probabilities)
            # find threshold closest to 0.5 for RF (different for other classifiers)
            close_default_rf = np.argmin(np.abs(threshold - 0.5))
            # compute ROC AUC
            roc_auc = auc(false_positive_rate, true_posistive_rate)
            # plot ROC curve
            plt.subplots(figsize=(7, 5.5))
            plt.plot(false_positive_rate, true_posistive_rate, color='darkorange', lw=2,
                     label='ROC curve (area = %0.2f)' % roc_auc)
            plt.plot([0, 1], [0, 1], color='navy', lw=2, linestyle='--')
            plt.plot(false_positive_rate[close_default_rf], true_posistive_rate[close_default_rf], 'o',
                     markersize=10, label="threshold 0.5RF", fillstyle="none", c='k', mew=2)
            plt.xlim([0.0, 1.0])
            plt.ylim([0.0, 1.05])
            plt.xlabel('False Positive Rate')
            plt.ylabel('True Positive Rate')
            plt.title('ROC Curve')
            plt.legend(loc="lower right")
            
        # def confusion_matrix_plot(self,figsize):
            
        #     # create confusion matrix
        #     matrix = confusion_matrix(self.self_obj.y_data, self.self_obj.X_hat_data["predict"].to_numpy())
        #     plt.figure(figsize=figsize)
        #     mglearn.tools.heatmap(matrix, xlabel='Predicted label', ylabel='True label',
        #                           xticklabels=np.array(pd.Series(self.self_obj.type_class).str.slice(0,2)),
        #                           yticklabels=self.self_obj.type_class, cmap=plt.cm.viridis_r, fmt="%d")
        #     plt.gca().invert_yaxis()
        def confusion_matrix_plot(self, figsize):
            # create confusion matrix
            matrix = confusion_matrix(self.self_obj.y_data, self.self_obj.X_hat_data["predict"].to_numpy())
            
            # Assuming type_class and x and y labels
            xticklabels = np.array(pd.Series(self.self_obj.type_class).str.slice(0,2))
            yticklabels = self.self_obj.type_class

            # Set figure size
            plt.figure(figsize=figsize)

            # Use seaborn
            sns.heatmap(matrix, annot=True, cmap='viridis_r', fmt='d', xticklabels=xticklabels, yticklabels=yticklabels)

            # Configure the plot details
            plt.gca().invert_yaxis()
            plt.xlabel('Predicted label')
            plt.ylabel('True label')
    
        def confusion_matrix_plot_percentage(self, figsize):
            # create confusion matrix
            matrix = confusion_matrix(self.self_obj.y_data, self.self_obj.X_hat_data["predict"].to_numpy())
            
            # Convert to percentage
            matrix = matrix.astype('float') / matrix.sum(axis=1)[:, np.newaxis]

            # Assuming type_class and x and y labels
            xticklabels = np.array(pd.Series(self.self_obj.type_class).str.slice(0,2))
            yticklabels = self.self_obj.type_class

            # Set figure size
            plt.figure(figsize=figsize)

            # Use seaborn
            sns.heatmap(matrix, annot=True, cmap='viridis_r', fmt='.2%', xticklabels=xticklabels, yticklabels=yticklabels)

            # Configure the plot details
            plt.gca().invert_yaxis()
            plt.xlabel('Predicted label')
            plt.ylabel('True label')
    

        def plot(self,save_path,file_name,figsize):
            
            if len(pd.unique(self.self_obj.y_data)) <= 2:
                self.two_features_roc()
                plt.title(file_name+'ROC curve')
                plt.savefig(save_path+file_name+'_roc.png',dpi=1000,bbox_inches="tight")
                plt.savefig(save_path+file_name+'_roc.pdf',dpi=1000,bbox_inches="tight")
                plt.close()
                self.confusion_matrix_plot(figsize)
                plt.title(file_name+'Confusion matrix')
                plt.savefig(save_path+file_name+'_confusion_matrix.png',dpi=1000,bbox_inches="tight")
                plt.savefig(save_path+file_name+'_confusion_matrix.pdf',dpi=1000,bbox_inches="tight")
                plt.close()
                self.confusion_matrix_plot_percentage(figsize)
                plt.title(file_name+'Confusion matrix percentage')
                plt.savefig(save_path+file_name+'_confusion_matrix_percentage.png',dpi=1000,bbox_inches="tight")
                plt.savefig(save_path+file_name+'_confusion_matrix_percentage.pdf',dpi=1000,bbox_inches="tight")
                plt.close()
                
            else:
                self.confusion_matrix_plot(figsize)
                plt.title(file_name+'Confusion matrix')
                plt.savefig(save_path+file_name+'_confusion_matrix.png',dpi=1000,bbox_inches="tight")
                plt.savefig(save_path+file_name+'_confusion_matrix.pdf',dpi=1000,bbox_inches="tight")
                plt.close()
                self.confusion_matrix_plot_percentage(figsize)
                plt.title(file_name+'Confusion matrix percentage')
                plt.savefig(save_path+file_name+'_confusion_matrix_percentage.png',dpi=1000,bbox_inches="tight")
                plt.savefig(save_path+file_name+'_confusion_matrix_percentage.pdf',dpi=1000,bbox_inches="tight")
                plt.close()
    @ClassificationEvaluatePlot
    def classification_evaluate_plot(self,save_path,file_name,figsize, *args, **kwargs):
        """
        Parameters
        ----------
        save_path : str
            input is file or plot save path way with no contain file name.
        file_name : str
            only indicate test or train but not specific file name, 
            such as input can't be classification_report.csv.
        figsize : tuple
            set the pdf size of output images
        *args : NoneType
            Placeholder parameter, no real meaning for now.
        **kwargs : NoneType
            Placeholder parameter, no real meaning for now.

        Returns
        -------
        Export the files of evaluation metrics such as f1-score of the classification 
        to the specified path.
        output : _classification_report.csv and _classification_report2.csv
        
        """
        pass
      